/** @file websocket.c
 *  @brief This module
 */

/***************************************************************************************************
 * Includes
 **************************************************************************************************/

#include "websocket.h"

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "ws.h"

#include "../../rpi_param.h"

/***************************************************************************************************
 * Preprocessor Constants and Macros
 **************************************************************************************************/

#define MAX_TOKEN 10

/***************************************************************************************************
 * Type and Contant Definitions
 **************************************************************************************************/

/***************************************************************************************************
 * Persistent Variables
 **************************************************************************************************/

int client_fd = -1;

/***************************************************************************************************
 * Internal Function Prototypes
 **************************************************************************************************/
/**
 * @brief Called when a client connects to the server.
 *
 * @param fd File Descriptor belonging to the client. The @p fd parameter
 * is used in order to send messages and retrieve informations
 * about the client.
 */
void onopen(int fd);

/**
 * @brief Called when a client disconnects to the server.
 *
 * @param fd File Descriptor belonging to the client. The @p fd parameter
 * is used in order to send messages and retrieve informations
 * about the client.
 */
void onclose(int fd);

/**
 * @brief Called when a client connects to the server.
 *
 * @param fd File Descriptor belonging to the client. The
 * @p fd parameter is used in order to send messages and
 * retrieve informations about the client.
 *
 * @param msg Received message, this message can be a text
 * or binary message.
 *
 * @param size Message size (in bytes).
 *
 * @param type Message type.
 */
void onmessage(int fd, const unsigned char *msg, uint64_t size, int type);

void mode_command_handler(char *payload);

void pattern_command_handler(char *payload);

void color_command_handler(char *payload);

/***************************************************************************************************
 * External Function Definitions
 **************************************************************************************************/

void websocket_start() {
  struct ws_events evs;
  evs.onopen = &onopen;
  evs.onclose = &onclose;
  evs.onmessage = &onmessage;
  ws_socket(&evs, 8080, 1);
}

void websocket_stop() {}

void websocket_send_mode(int mode) {
  char msg[] = "m:0";
  msg[2] = '0' + mode;
  if (client_fd != -1) {
    log_debug("Sending mode \"%s\"", msg);
    ws_sendframe_txt(client_fd, msg, true);
  }
}

/***************************************************************************************************
 * Internal Function Definitions
 **************************************************************************************************/

void onopen(int fd) {
  char *cli;
  cli = ws_getaddress(fd);
  log_debug("Connection opened, client: %d | addr: %s", fd, cli);
  free(cli);
  client_fd = fd;
}

void onclose(int fd) {
  char *cli;
  cli = ws_getaddress(fd);
  log_debug("Connection closed, client: %d | addr: %s", fd, cli);
  free(cli);
  client_fd = -1;
}

void onmessage(int fd, const unsigned char *msg, uint64_t size, int type) {
  char msgStr[size + 1];
  char *command, *payload;
  char *cli = ws_getaddress(fd);
  log_trace("Received message: %s (size: %" PRId64 ", type: %d), from: %s/%d", msg, size, type, cli,
            fd);

  strcpy(msgStr, (char *)msg);
  command = strtok(msgStr, ":");
  if (command != NULL) {
    payload = strtok(NULL, ":");
    if (payload != NULL && strlen(command) == 1) {
      switch (command[0]) {
      case 'm':
        mode_command_handler(payload);
        break;
      case 'p':
        pattern_command_handler(payload);
        break;
      case 'c':
        color_command_handler(payload);
        break;

      default:
        log_warn("Unkown command in \"%s\"", msgStr);
        break;
      }
    } else {
      log_warn("Empty payload in \"%s\"", msgStr);
    }
  } else {
    log_warn("No command found in \"%s\"", msgStr);
  }

  free(cli);
}

void mode_command_handler(char *payload) {
  int newMode = atoi(payload);
  if (0 <= newMode && newMode <= 3) {
    set_mode(newMode);
  } else {
    log_warn("Unknown mode : %s", payload);
  }
}

void pattern_command_handler(char *payload) {
  int newPattern = atoi(payload);
  if (0 <= newPattern && newPattern <= 3) {
    set_pattern(newPattern);
  } else {
    log_warn("Unknown pattern : %s", payload);
  }
}

void color_command_handler(char *payload) {
  int i = 0;
  char *tokenArray[MAX_TOKEN] = {NULL};
  tokenArray[i] = strtok(payload, ",");
  while (tokenArray[i] && i < MAX_TOKEN - 1) {
    tokenArray[++i] = strtok(NULL, ",");
  }
  log_debug("tokens[%d]", i);
  for (size_t n = 0; n < i; ++n) {
    log_debug(" - %s", tokenArray[n]);
  }
  if (atoi(tokenArray[0]) == 0) {
    set_hue_base(LED_NCHANS, atoi(tokenArray[1]) * INT8_MAX / HUE_MAX);
    set_saturation(LED_NCHANS, atoi(tokenArray[2]) * INT8_MAX / 100);
    set_luminosity(LED_NCHANS, atoi(tokenArray[3]) * INT8_MAX / 100);
  }
}